import pytest

from clicast.filters import match_cli_args, match_program_or_subcommand


def test_match_cli_args():
  msg_text = 'Message for -b option\nLine 2'
  msg = r'[ -b \w+] %s' % msg_text

  assert match_cli_args(msg, cli_args='./cli-command -b bug -i issue') == msg_text
  assert match_cli_args(msg, cli_args='./cli-command -i issue') is None


@pytest.mark.parametrize('msg', ['[program] Message', '[sc|subcommand] Message'])
def test_match_program_and_sub_command(msg):
  assert match_program_or_subcommand(msg, cli_args='./program subcommand') == 'Message'
  assert match_program_or_subcommand(msg, cli_args='./program sc') == 'Message'
  assert match_program_or_subcommand(msg, cli_args='./anotherprogram prefixsubcommand -p program -s subcommand') is None
  assert match_program_or_subcommand(msg, cli_args='./anotherprogram scsuffix -p program -s subcommand') is None



import os
import sys

from clicast.cast import Cast, CastReader


def check_message(url, msg_filter=None, cache_duration=None, allow_exit=False, raises=False, local_file=None, reset=False, **show_kwargs):
  """
  Check remote url for new messages and display them.

  :param str url: Cast file URL to check
  :param callable msg_filter: Filter messages with callable that accepts message string and alert boolean (True for
                              alert message). It should return the original or an updated message, or None if the
                              message should be ignored.
  :param int cache_duration: Cache messages locally for number of seconds to avoid checking the URL too often.
                             This is useful for response latency sensitive CLI to ensure user's experience
                             isn't compromised. Alternatively, you may want to check messages in a seperate thread.
  :param bool allow_exit: Perform sys.exit(1) if cast requests it.
  :param bool raises: Raise exception for failed to download/parse cast file or such.
                      Recommended to set this to False for production / in non-debug mode.
  :param str local_file: Local file path for the cast file URL. It will be used instead of 'url' if exists.
                         This is primarily used for testing / in development.
  :param bool reset: Reset read messages. All messages will be displayed again like the user runs this for the first
                     time
  :param dict show_kwargs: kwargs to be passed to :meth:`clicast.CastReader.show_messages`
  """

  try:
    if local_file and os.path.exists(local_file):
      cast = Cast.from_content(local_file, msg_filter)
    else:
      cast = Cast.from_content(url, msg_filter, cache_duration)

    if reset:
      CastReader.reset()

    reader = CastReader(cast)
    reader.show_messages(**show_kwargs)

    if allow_exit and cast.alert_exit:
      sys.exit(1)
  except Exception:
    if raises:
      raise



import os
import tempfile

from clicast.cast import Cast, CastReader


CAST_URL = 'https://raw.githubusercontent.com/maxzheng/clicast/master/tests/example.cast'
CAST_FILE = os.path.join(os.path.dirname(__file__), 'example.cast')


class TestCast(object):
  def test_from_file(self):
    cast = Cast.from_content(CAST_FILE)

    assert cast.alert == 'We found a big bad bug. Please try not to step on it!! Icky...\nNo worries. It will be fixed soon! :)'
    assert cast.alert_exit
    assert [m.message for m in cast.messages] == [
      'Version 0.1 has been released! If you upgrade, you will get:\n'
      '1) Cool feature 1\n'
      '2) Cool feature 2\n'
      'So what are you waiting for? :)',
      'Version 0.2 has been released! Upgrade today to get cool features.',
      'There is a small bug over there, so watch out!',
      '[-f\\b] A bug that affects the -f option. (applies only if `clicast.filters.match_cli_args` filter is used)'
    ]

  def test_save(self):
    from_content = open(CAST_FILE).read()
    cast = Cast.from_content(CAST_FILE)

    to_cast_file = os.path.join(tempfile.gettempdir(), 'clicast.to_file_test.cast')
    try:
      cast.save(to_cast_file)
      to_content = open(to_cast_file).read()
      assert from_content == to_content
    finally:
      if os.path.exists(to_cast_file):
        os.unlink(to_cast_file)

  def test_from_url(self):
    cast = Cast.from_content(CAST_URL)
    assert cast.messages

    from remoteconfig import RemoteConfig
    c = RemoteConfig(CAST_URL)
    assert str(c)

  def test_add_msg(self):
    cast = Cast()
    cast.add_msg('Message 1')
    cast.add_msg('Message Alert', alert=True)
    cast.add_msg('Message 2')

    assert cast.alert == 'Message Alert'
    assert not cast.alert_exit
    assert [(m.key, m.message) for m in cast.messages] == [
      ('1', 'Message 1'),
      ('2', 'Message 2')
    ]

    cast.add_msg('Message Alert Exit', alert_exit=True)

    assert cast.alert == 'Message Alert Exit'
    assert cast.alert_exit

  def test_del_msg(self):
    cast = Cast()
    cast.add_msg('Message 1')
    cast.add_msg('Message 2')
    cast.add_msg('Message Alert', alert_exit=True)
    cast.del_msg()

    assert cast.alert == 'Message Alert'
    assert cast.alert_exit
    assert [(m.key, m.message) for m in cast.messages] == [('2', 'Message 2')]

    del_count = cast.del_msg(100)
    assert del_count == 1

    cast.add_msg('Message 3')
    cast.add_msg('Message 4')
    cast.add_msg('Message 5')
    cast.del_msg(2)
    cast.del_msg(alert=True)

    assert not cast.alert
    assert not cast.alert_exit
    assert [(m.key, m.message) for m in cast.messages] == [('5', 'Message 5')]

    cast_file = os.path.join(tempfile.gettempdir(), 'clicast.to_file_test.cast')
    try:
      cast.save(cast_file)
      cast = Cast.from_content(cast_file)
      cast.del_msg(100)

      cast.save(cast_file)
      cast = Cast.from_content(cast_file)
      cast.add_msg('Message 6')

      assert str(cast) == '[Messages]\n6: Message 6'
    finally:
      if os.path.exists(cast_file):
        os.unlink(cast_file)

  def test_set_msg_limit(self):
    cast = Cast()
    cast.set_msg_limit(2)
    for x in range(10):
      cast.add_msg('Message %d' % x)
    assert str(cast) == '[Messages]\n9: Message 8\n10: Message 9\n_limit: 2'

  def test_filter(self):
    def msg_filter(msg, alert=False):
      if 'small bug' in msg:
        return msg
    cast = Cast.from_content(CAST_FILE, msg_filter)
    assert str(cast) == '[Messages]\n3: There is a small bug over there, so watch out!\n_limit: 5'


class TestCastReader(object):
  def setup_class(cls):
    CastReader.READ_MSG_FILE = '/tmp/clicast.test.read'
    if os.path.exists(CastReader.READ_MSG_FILE):
      os.unlink(CastReader.READ_MSG_FILE)

  def test_new_messages(self):
    cast = Cast.from_content(CAST_FILE)

    reader = CastReader(cast)
    assert reader.new_messages() == [
      'We found a big bad bug. Please try not to step on it!! Icky...\nNo worries. It will be fixed soon! :)',
      'Version 0.1 has been released! If you upgrade, you will get:\n'
      '1) Cool feature 1\n'
      '2) Cool feature 2\n'
      'So what are you waiting for? :)',
      'Version 0.2 has been released! Upgrade today to get cool features.',
      'There is a small bug over there, so watch out!',
      '[-f\\b] A bug that affects the -f option. (applies only if `clicast.filters.match_cli_args` filter is used)'
    ]



"""
An utility script to help create or edit cast files.

Example Action                | Run Command
==============================================================
View current cast             | cast
--------------------------------------------------------------
Add a new message to new file | cast "New Message" -f new.cast
To delete a message           | cast -d
--------------------------------------------------------------
Set an alert message          | cast "New Alert" -a
Unset an alert                | cast -d -a
--------------------------------------------------------------
"""

import argparse
from glob import glob
import logging
import os
import sys

from clicast.cast import Cast


logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')
log = logging.getLogger(__name__)


def cast():
  parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
  parser.add_argument('-a', '--alert', action='store_true', help='Indicates this is an alert message')
  parser.add_argument('-e', '--alert-exit', action='store_true', help='Indicates this is an alert message with exit = True')
  parser.add_argument('-f', '--file', help='New or existing cast file to update. Defaults to any cast file in current directory.')
  group = parser.add_mutually_exclusive_group()
  group.add_argument('msg', nargs='?', help='The message to cast')
  group.add_argument('--delete', type=int, metavar='COUNT', nargs='?', const=1,
                     help='Delete the oldest message (default) or the number of messages (oldest first). '
                          'Use with --alert to remove the alert message')
  group.add_argument('--limit', type=int, metavar='COUNT',
                     help='Set limit on total number of messages by deleting oldest message as new message is added '
                          'when limit has been reached.')

  args = parser.parse_args()

  cast, cast_file = cast_info_from_path(args.file)

  if args.delete:
    cast.del_msg(args.delete, args.alert or args.alert_exit)

  elif args.limit:
    cast.set_msg_limit(args.limit)

  elif args.msg:
    cast.add_msg(args.msg, args.alert, args.alert_exit)

  elif not os.path.exists(cast_file):
    log.error('%s does not exist', cast_file)
    sys.exit(1)

  cast.save(cast_file)

  print(str(cast).strip())


def cast_info_from_path(cast_file=None):
  if cast_file:
    if os.path.exists(cast_file):
      cast = Cast.from_content(cast_file)
    else:
      cast = Cast()

  else:
    cast_files = glob('*.cast')

    if not cast_files:
      log.error('There is no cast file. To create a new one, specify the name with --file')
      sys.exit(1)

    if len(cast_files) > 1:
      log.error('There are more than one cast file. Please specify which to modify with --file')
      sys.exit(1)

    cast_file = cast_files[0]
    cast = Cast.from_content(cast_file)

  return cast, cast_file



from functools import total_ordering
import os
import re
import sys
import tempfile

from remoteconfig import RemoteConfig


class CastError(Exception):
  pass


class Cast(object):
  """ Represents all cast options/messages. """

  ALERT_SECTION = 'Alert'
  MESSAGES_SECTION = 'Messages'
  ALERT_MSG_KEY = 'message'
  ALERT_EXIT_KEY = 'exit'
  MESSAGE_NEXT_KEY = '_next_key'
  MESSAGE_LIMIT_KEY = '_limit'

  @total_ordering
  class CastMessage(object):
    """ Represents a single message in a cast. """
    def __init__(self, key, message):
      """
      :param key: Message key
      :param str message: The actual message
      """
      self.key = str(key)
      self.message = message

    def __eq__(self, other):
      return self.key == other.key

    def __lt__(self, other):
      try:
        return int(self.key) < int(other.key)
      except Exception:
        return self.key < other.key

  def __init__(self, alert=None, alert_exit=False, messages=None, next_msg_key=None, msg_limit=None):
    """
    :param str alert: Alert message
    :param bool alert_exit: Should client CLI exit. Ignored unless alert message is set.
    :param list(tuple) messages: List of tuple of (key, message)
    :param int next_msg_key: Next message key to use
    :param int msg_limit: Limit total number of messages by deleting oldest one.
    """
    self.alert = alert
    self.alert_exit = alert and alert_exit
    self.messages = messages and sorted([self.CastMessage(*m) for m in messages]) or []
    self._next_msg_key = next_msg_key
    self.msg_limit = msg_limit

    # Always set this so that it can be used in :meth:`self.save`
    if self.messages and not self._next_msg_key:
      self._next_msg_key = self.next_msg_key(reserve_next=False)

  def add_msg(self, msg, alert=False, alert_exit=False):
    """
    :param str msg: The message to add or set
    :param bool alert: Indicates this is the alert message to set.
    :param bool alert_exit: Indicates this is the alert should request client to exit.
    """
    if alert or alert_exit:
      self.alert = msg
      self.alert_exit = alert_exit
    else:
      self.messages.append(self.CastMessage(self.next_msg_key(), msg))
      self.set_msg_limit(self.msg_limit)

  def del_msg(self, count=1, alert=False):
    """
    Delete message

    :param int count: Number of messages to delete
    :param bool alert: Delete the alert message. Param `count` is ignored.
    """
    if alert:
      self.alert = None
      self.alert_exit = None
      return 1
    else:
      before_count = len(self.messages)
      self.messages = self.messages[count:]
      return before_count - len(self.messages)

  def next_msg_key(self, reserve_next=True):
    """
      :param bool reserve_next: Indicates the key after next should be reserved. Default behavior.
      :return: Next message key
      :rtype: int
    """
    if not self._next_msg_key:
      keys = []

      for m in self.messages:
        try:
          keys.append(int(m.key))
        except Exception:
          pass

      if keys:
        self._next_msg_key = keys[-1] + 1
      else:
        self._next_msg_key = 1

    next_key = self._next_msg_key

    if reserve_next:
      self._next_msg_key += 1

    return next_key

  def set_msg_limit(self, limit=None):
    """
    Limit total number of messages by deleting oldest message as new message is added when limit has been reached.

    :param int limit: Number of message to limit. Defaults to infinite.
    """
    self.msg_limit = limit

    if limit:
      over_limit = len(self.messages) - limit
      if over_limit > 0:
        self.del_msg(over_limit)

  @classmethod
  def from_content(cls, context, msg_filter=None, cache_duration=None):
    """ Create a :class:`Cast` from the given content

    :param str context: Cast content string, file, or URL
    :param callable msg_filter: Filter messages with callable that accepts message string and alert boolean (True for
                                alert message). It should return the original or an updated message, or None if the
                                message should be ignored.
    :param int cache_duration: For cast URL only. Optionally cache the URL content for the given duration (seconds) to
                               avoid downloading too often.
    :rtype: :class:`Cast`
    """
    config = RemoteConfig(context, cache_duration=cache_duration, compact_form=True)

    alert_msg = None
    alert_exit = None

    if cls.ALERT_SECTION in config:
      for key, value in config.items(cls.ALERT_SECTION):
        if cls.ALERT_MSG_KEY == key:
          if msg_filter:
            value = msg_filter(value, True)
          alert_msg = value
        elif cls.ALERT_EXIT_KEY == key:
          alert_exit = value
        else:
          raise CastError('Invalid key "%s" in %s section', key, cls.ALERT_SECTION)

    messages = []
    next_msg_key = None
    msg_limit = None

    if cls.MESSAGES_SECTION in config:
      for key, value in config.items(cls.MESSAGES_SECTION):
        if key == cls.MESSAGE_NEXT_KEY:
          next_msg_key = value
        elif key == cls.MESSAGE_LIMIT_KEY:
          msg_limit = value
        elif key.startswith('_'):
          pass  # Ignore future private keys
        else:
          if msg_filter:
            value = msg_filter(value)
          if value:
            messages.append((key, value))

    return cls(alert_msg, alert_exit, messages, next_msg_key, msg_limit)

  def __repr__(self):
    parser = RemoteConfig(kv_sep=': ', compact_form=True)

    if self.alert:
      parser.add_section(self.ALERT_SECTION)
      parser.set(self.ALERT_SECTION, self.ALERT_MSG_KEY, self.alert)
      if self.alert_exit:
        parser.set(self.ALERT_SECTION, self.ALERT_EXIT_KEY, True)

    if self.messages:
      parser.add_section(self.MESSAGES_SECTION)
      for msg in self.messages:
        parser.set(self.MESSAGES_SECTION, msg.key, msg.message)

    elif self._next_msg_key:
      parser.add_section(self.MESSAGES_SECTION)
      parser.set(self.MESSAGES_SECTION, self.MESSAGE_NEXT_KEY, self._next_msg_key)

    if self.msg_limit:
      parser.set(self.MESSAGES_SECTION, self.MESSAGE_LIMIT_KEY, self.msg_limit)

    parser._indent_spaces = len(str(self._next_msg_key)) + 2 if self._next_msg_key else 3

    return str(parser).strip()

  def save(self, cast_file):
    """ Save the cast data to the given file. """
    with open(cast_file, 'w') as fp:
      fp.write(str(self) + '\n')


class CastReader(object):
  """ Reads a :class:`Cast` and keep track of read messages """

  READ_MSG_FILE = os.path.join(tempfile.gettempdir(), '%s.read_messages' % os.path.basename(sys.argv[0]))

  @classmethod
  def reset(cls):
    """ Resets read messages, so all messages will be displayed again. """
    if os.path.exists(cls.READ_MSG_FILE):
      os.unlink(cls.READ_MSG_FILE)

  def __init__(self, cast):
    """
    :param Cast cast: A :class:`Cast` instance
    """
    self.cast = cast

  def show_messages(self, logger=None, header=None, footer=None):
    """
    Show new messages

    :param logging.Logger logger: Use logger to print the new messages instead of stdout
    :param str header: Header to show before messages
    :param str footer: Footer to show before messages
    """
    msgs = self.new_messages()

    if msgs:
      if logger:
        if header:
          logger.info(header)
        for msg in msgs:
          for line in msg.split('\n'):
            logger.info(line)
        if footer:
          logger.info(footer)

      else:
        if header:
          print(header)
        print('\n\n'.join(msgs))
        if footer:
          print(footer)

  def new_messages(self, mark_as_read=True):
    """
    :param bool mark_as_read: Mark new messages as read
    :return: List of new messages with alert being the first if any.
    :rtype: list(str)
    """
    read_keys = self._read_msg_keys()
    new_messages = [m for m in self.cast.messages if m.key not in read_keys]

    if new_messages and mark_as_read:
      self._mark_as_read(new_messages)

    msgs = [m.message for m in new_messages]

    if self.cast.alert:
      msgs.insert(0, self.cast.alert)

    return msgs

  def _read_msg_keys(self):
    """ Set of read messages. """

    try:
      with open(self.READ_MSG_FILE) as fp:
        read_keys = fp.read()
        return set(read_keys.split())
    except Exception:
      return set()

  def _mark_as_read(self, messages):
    """ Mark the given list of :class:`CastMessage` as read. """

    keys = self._read_msg_keys()
    keys.update(m.key for m in messages)

    with open(self.READ_MSG_FILE, 'w') as fp:
      fp.write(' '.join(keys))


def _re_sub_multiline(pattern, repl, string):
  """ Simple hack to get multiline working in Python 2.6 and higher """
  try:
    content = re.sub(pattern, repl, string, flags=re.MULTILINE)
  except Exception:
    content = []
    for line in string.split('\n'):
      content.append(re.sub(pattern, repl, line))
    content = '\n'.join(content)

  return content



import pytest
import subprocess


@pytest.mark.parametrize("script", ['bin/cast-example'])
def test_script(script):

  try:
    subprocess.check_output([script, '-h'], stderr=subprocess.STDOUT)
  except subprocess.CalledProcessError as e:
    print(e.output)
    assert e.returncode == 0



#!/usr/bin/env python2.6

import setuptools


setuptools.setup(
    name='clicast',
    version='1.0.0',

    author='Max Zheng',
    author_email='maxzheng.os @t gmail.com',

    description='Broadcast messages for CLI tools, such as a warning for critical bug '
                'or notification about new features.',
    long_description=open('README.rst').read(),

    url='https://github.com/maxzheng/clicast',

    entry_points={
        'console_scripts': [
            'cast = clicast.editor:cast',
        ],
    },

    install_requires=open('requirements.txt').read(),

    license='MIT',

    packages=setuptools.find_packages(),
    include_package_data=True,

    setup_requires=['setuptools-git', 'wheel'],

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Topic :: Software Development',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],

    keywords='cli broadcast command warning critical bug',
)



import os
import re
import sys


REGEX_IN_MSG_RE = re.compile(r'^\[(.+)\] *(.+)$', flags=re.DOTALL)
IS_WORD_RE = re.compile(r'^\w+$')


def match_program_or_subcommand(msg, alert=False, cli_args=None):
  """
  Look for messages that starts with "[program|sub-command] message" and search the program or sub-command against
  the the first two words in the given args (or in sys.argv).

  :param str msg: Message to search
  :param bool alert: Is alert message?
  :param list/str cli_args: Optional args used for testing instead of sys.argv
  :ret str: New message without [program|sub-command] or None if program or sub-command doesn't match CLI args
  """
  match = REGEX_IN_MSG_RE.match(msg)

  if match:
    regex = match.group(1)
    msg = match.group(2)

    if not (regex.startswith('^') or regex.endswith('$')):
      regex = '^(?:%s)$' % regex

    if isinstance(cli_args, str):
      cli_args = cli_args.split()
    elif not cli_args:
      cli_args = sys.argv

    commands = [os.path.basename(cli_args[0])]
    if len(cli_args) > 1 and IS_WORD_RE.match(cli_args[1]):
      commands.append(cli_args[1])

    match = any(re.search(regex, command) for command in commands)
    if not match:
      return None

  return msg


def match_cli_args(msg, alert=False, cli_args=None):
  """
  Look for messages that starts with "[pattern] message" and search the pattern against
  the given args (or sys.argv).

  :param str msg: Message to search
  :param bool alert: Is alert message?
  :param list cli_args: Optional args used for testing instead of sys.argv
  :ret str: New message without [pattern] or None if pattern doesn't match CLI args
  """
  match = REGEX_IN_MSG_RE.match(msg)

  if match:
    regex = match.group(1)
    msg = match.group(2)
    if isinstance(cli_args, str):
      cli_args = cli_args.split()
    args = ' '.join(cli_args or sys.argv)
    if not re.search(regex, args):
      return None

  return msg
